/**
 ****************************************************************************************************
 * @file        chflash.c
 * @author      ԭŶ(ALIENTEK)
 * @version     V1.0
 * @date        2023-07-20
 * @brief       CH32ڲFLASHд 
 * @license     Copyright (c) 2020-2032, ӿƼ޹˾
 ****************************************************************************************************
 * @attention
 *
 * ʵƽ̨:ԭ CH32V307
 * Ƶ:www.yuanzige.com
 * ̳:www.openedv.com
 * ˾ַ:www.alientek.com
 * ַ:openedv.taobao.com
 *
 * ޸˵
 * V1.0 20230720
 * һη
 *
 ****************************************************************************************************
 */

#include "./SYSTEM/usart/usart.h"
#include "./SYSTEM/delay/delay.h"
#include "./BSP/CHFLASH/chflash.h"

/**
 * @brief       ַָȡһ (16λ)
 * @param       faddr   : ȡַ (˵ַΪ2ı!!)
 * @retval      ȡ (16λ)
 */
uint16_t chflash_read_halfword(uint32_t faddr)
{
    return *(volatile uint16_t *)faddr;
}

/**
 * @brief       ַָʼָȵ
 * @param       raddr : ʼַ
 * @param       pbuf  : ָ
 * @param       length: Ҫȡİ(16λ),2ֽڵ
 * @retval      
 */
void chflash_read(uint32_t raddr, uint16_t *pbuf, uint16_t length)
{
    uint16_t i;

    for (i = 0; i < length; i++)
    {
        pbuf[i] = chflash_read_halfword(raddr); /* ȡ2ֽ */
        raddr += 2;                             /* ƫ2ֽ */
    }
}

/**
 * @brief       д
 * @note        Ѿԭҳ
 * @param       waddr   : ʼַ (˵ַΪ2ı!!д!)
 * @param       pbuf    : ָ
 * @param       length  : Ҫд (16λ)
 * @retval      
 */
void chflash_write_nocheck(uint32_t waddr, uint16_t *pbuf, uint16_t length)
{
    uint16_t i;
    for (i = 0; i < length; i++)
    {
        FLASH_ProgramHalfWord(waddr, pbuf[i]);
        waddr += 2;                             /* ָһ */
    }
}

uint16_t g_flashbuf[CH32_SECTOR_SIZE / 2];      /* ȡݵBUF */

/**
 * @brief       FLASHָλ,дָȵ(Զ)
 *   @note      úCH32ڲFLASHָλдָȵ
 *              úȼҪдҳǷΪ(ȫ0XFFFF)
 *              , Ȳǣֱҳд.
 * @param       waddr   : ʼַ (˵ַΪ2ı!! д!)
 * @param       pbuf    : ָ
 * @param       length  : Ҫд (16λ)
 * @retval      
 */
void chflash_write(uint32_t waddr, uint16_t *pbuf, uint16_t length)
{
    uint32_t secpos;        /* ҳַ */
    uint16_t secoff;        /* ҳƫƵַ(16λּ) */
    uint16_t secremain;     /* ҳʣַ(16λּ) */
    uint16_t i;
    uint32_t offaddr;       /* ȥ0X08000000ĵַ */

    if (waddr < CH32_FLASH_BASE || waddr % 2 || (waddr >= (CH32_FLASH_BASE + CH32_FLASH_SIZE)))
    {
        return;             /* Ƿַ */
    }

    FLASH_Unlock();         /* FLASH */

    offaddr = waddr - CH32_FLASH_BASE;              /* ʵƫƵַ */
    secpos = offaddr / CH32_SECTOR_SIZE;            /* ҳַ */
    secoff = (offaddr % CH32_SECTOR_SIZE) / 2;      /* ҳڵƫ(2ֽΪλ) */
    secremain = CH32_SECTOR_SIZE / 2 - secoff;      /* ҳʣռС */
    if (length <= secremain)
    {
        secremain = length;                         /* ڸҳΧ */
    }

    while (1)
    {
        chflash_read(secpos * CH32_SECTOR_SIZE + CH32_FLASH_BASE, g_flashbuf, CH32_SECTOR_SIZE / 2); /* ҳ */
        for (i = 0; i < secremain; i++)                                                              /* У */
        {
            if (g_flashbuf[secoff + i] != 0XFFFF)
            {
                break;                  /* Ҫ */
            }
        }
        if (i < secremain)              /* Ҫ */
        {
            FLASH_ErasePage(secpos * CH32_SECTOR_SIZE * 16 + CH32_FLASH_BASE);  /* FLASHһҳΪ256ֽڣǵ4K(256*16)ִֽв */

            for (i = 0; i < secremain; i++)                 /*  */
            {
                g_flashbuf[i + secoff] = pbuf[i];
            }

            chflash_write_nocheck(secpos * CH32_SECTOR_SIZE + CH32_FLASH_BASE, g_flashbuf, CH32_SECTOR_SIZE / 2); /* дҳ */
        }
        else
        {
            chflash_write_nocheck(waddr, pbuf, secremain);  /* ֱдҳʣռ */
        }
        if (length == secremain)
        {
            break;                      /* д */
        }
        else                            /* дδ */
        {
            secpos++;                   /* ҳַ1 */
            secoff = 0;                 /* ƫλΪ0 */
            pbuf += secremain;          /* ָƫ */
            waddr += secremain * 2;     /* дַƫ(16λݵַ,Ҫ*2) */
            length -= secremain;        /* ֽ(16λ)ݼ */

            if (length > (CH32_SECTOR_SIZE / 2))
            {
                secremain = CH32_SECTOR_SIZE / 2;           /* һҳд */
            }
            else
            {
                secremain = length;     /* һҳд */
            }
        }
    }

    FLASH_Lock();                       /*  */
}

/**
 * @brief       FLASHָλ, дָȵ(ֱӲ64Kڴ)
 *   @note      ú׷ٶȣУݣֱӿд룬д
 *              ûԼѡǷڴ棬ÿβڴ̶Ϊ64Kֽڣ˱ƫƵĵַ!!
 *              ⣬ģʽ£ÿдСλ1256ֽڣ
 * @param       waddr   : ʼַ (˵ַΪ2ı!!д!)
 * @param       pbuf    : ָ
 * @param       length  : Ҫд (16λ)
 * @param       erase   : 0,ڴ; 1,64Kڴ
 * @retval      
 */
void chflash_fast_write(uint32_t waddr, uint16_t *pbuf, uint16_t length, uint8_t erase)
{
    uint32_t secpos;        /* ҳַ */
    uint16_t secoff;        /* ҳƫƵַ(16λּ) */
    uint16_t secremain;     /* ҳʣַ(16λּ) */
    uint32_t offaddr;       /* ȥ0X08000000ĵַ */

    if (waddr < CH32_FLASH_BASE || waddr % 2 || (waddr >= (CH32_FLASH_BASE + CH32_FLASH_SIZE)))
    {
        return;             /* Ƿַ */
    }

    FLASH_Unlock_Fast();

    offaddr = waddr - CH32_FLASH_BASE;              /* ʵƫƵַ */
    secpos = offaddr / CH32_SECTOR_SIZE;            /* ҳַ */
    secoff = (offaddr % CH32_SECTOR_SIZE) / 2;      /* ҳڵƫ(2ֽΪλ) */
    secremain = CH32_SECTOR_SIZE / 2 - secoff;      /* ҳʣռС */

    if (length <= secremain)
    {
        secremain = length;                         /* ڸҳΧֱһд */
    }

    if (erase == 1)                                 /* Ҫ */
    {
        FLASH_EraseBlock_64K_Fast(waddr);           /* ֱӲ64Kֽڵڴ棬Ϊд׼ */
    }

    while (1)
    {
        /* д */
        FLASH_ProgramPage_Fast(secpos * CH32_SECTOR_SIZE + CH32_FLASH_BASE, (uint32_t*)pbuf);

        if (length == secremain)
        {
            break;                                  /* д */
        }
        else                                        /* дδ */
        {
            secpos++;                               /* ҳַ1 */
            pbuf += secremain;                      /* ָƫ */
            length -= secremain;                    /* ֽ(16λ)ݼ */

            if (length > (CH32_SECTOR_SIZE / 2))
            {
                secremain = CH32_SECTOR_SIZE / 2;   /* һҳд */
            }
            else
            {
                secremain = length;                 /* һҳд */
            }
        }
    }
    FLASH_Lock_Fast();
}

/**
 * @brief       д(д1)
 * @note        עдĵַΧҴ˵ַΪ2ıҪд룡
 * @param       waddr : ʼַ
 * @param       wdata : Ҫд
 */
void test_write(uint32_t waddr, uint16_t wdata)
{
    chflash_write(waddr, &wdata, 1);    /* дһ */
}
